(* $Id: godi_plan.ml 117 2003-08-31 17:21:35Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_versions
open Godi_query
open Printf

exception Mission_impossible of string

module StrMap = Map.Make(String);;

module StrSet = Set.Make(String);;


let cat_weight =
  function
      `Installed_pkg _ -> 1
    | `Binary_pkg _ -> 2
    | `Source_pkg _ -> 3
;;


let create_pkg_hash pkgs =
  (* Creates a hash table mapping package names to lists of packages with
   * that name. The lists are sorted; the best package comes first.
   *)
  let ht = Hashtbl.create 100 in
  List.iter
    (fun pkg ->
       let name = (any_pkg pkg) # name in
       let old_list = try Hashtbl.find ht name with Not_found -> [] in
       Hashtbl.replace ht name (pkg :: old_list)
    )
    pkgs;
  (* Sort lists by version numbers, etc. *)
  Hashtbl.iter
    (fun name pkg_list ->
       let pkg_list' =
	 List.sort 
	   (fun pkg1 pkg2 ->
	      (* First sort by package type, then by version: *)
	      let w1 = cat_weight pkg1 in
	      let w2 = cat_weight pkg2 in
	      match w1 - w2 with
		  0 -> 
		    let v1 = (any_pkg pkg1)#version in
		    let v2 = (any_pkg pkg2)#version in
		    -(cmp_version_string v1 v2)
		| _ -> w1 - w2
	   )
	   pkg_list in
       Hashtbl.replace ht name pkg_list'
    )
    ht;
  ht
;;


let find_pkg ht name dep_list =
  (* Find the packages [name] that matches all conditions of [dep_list]: *)
  let candidates = 
    try Hashtbl.find ht name with Not_found -> [] in
  List.filter
    (fun pkg ->
       List.for_all
         (fun (name',cond) -> name'=name && matches_dep pkg (name, cond))
         dep_list)
    candidates
;;


exception Dead_end;;


let rec try_in_order ?(onerror=fun () -> ()) f l =
  match l with
      x :: l' ->
	(try f x 
	 with Dead_end -> 
	   try_in_order f l')
    | [] ->
	onerror();
	raise Dead_end
;;


let rec accumulate_deps f set deps followup_action =
  match deps with
      dep :: deps' ->
	let (name,cond) = dep in
	f 
	  set 
	  name 
	  dep 
	  (fun set' -> accumulate_deps f set' deps' followup_action)
    | [] ->
	followup_action set
;;


let get_requirements ~available ~selected () =
  let avail_ht = create_pkg_hash available in
  let sel_ht = create_pkg_hash selected in
  let msg = ref None in

  (* add_package: This is a backtracking algorithm trying to find the
   * best package match. It adds the best package [name] that has been
   * selected because of the dependency [dep] to the already selected
   * [set], and, if this is successful, calls [followup_action] with the
   * new set. The [followup_action] may raise Dead_end to signal that
   * this did not lead to a good end. In this case, the second-best package
   * [name] is selected instead, and so on, until a match is found.
   *
   * The sets are StrMap.t values (i.e. maps in reality) mapping names
   * to the pairs (pkg,deps), where pkg is the package and deps is the
   * list of collected dependencies that must all hold.
   *
   * The packages of [selected] are preferred over the packages of 
   * [available].
   *
   * I hope nobody gets really into the complexity of this algorithm...
   * I am sure there is a better one.
   *)

  let error m =
    match !msg with
	Some _ -> ()
      | None -> msg := Some m
  in

  let rec add_package set name dep (followup_action : 'a -> 'a) =
    let member = StrMap.mem name set in
    if member then (
      (* Case: the package is already selected. We only check whether
       * this package also fulfills the new condition dep. If not,
       * raise Dead_end. If yes, add the new conditions to the set, and
       * continue with the followup_action.
       *)
      let (pkg, stored_deps) = StrMap.find name set in
      if matches_dep pkg dep then (
	let set' = StrMap.add name (pkg, dep :: stored_deps) set in
	followup_action set'
      )
      else (
	error(sprintf
		"Version conflict for package %s:\n  %s"
		name
		(String.concat "\n  "
		   (List.map
		      print_dependency
		      (dep :: stored_deps))));
	raise Dead_end
      )
    )
    else (
      (* Case: the package is unknown. Get the list of possible packages
       * that fulfill dep, and try them in turn, until a path is found.
       * Every package has its own dependencies which are tried as
       * a followup_action.
       *)
      (* First try [selected], then [available] *)
      let pkg_list1 =
	find_pkg sel_ht name [dep] in
      let pkg_list =
	if pkg_list1 = [] then
	  find_pkg avail_ht name [dep]
	else
	  pkg_list1 in
      (* The best package is the first in pkg_list (because the packages
       * have been sorted by create_pkg_hash).
       *)
      if pkg_list = [] then
	error(sprintf "Package not found: %s" (print_dependency dep));
      try_in_order
	(fun pkg ->
	   (* Try pkg *)
	   let set' =
	     StrMap.add name (pkg, [dep]) set in
	   (* Add the dependencies in turn, finally do followup_action: *)
	   let deps' = (any_pkg pkg)#dependencies @
		       ( match pkg with
			     `Source_pkg _ -> 
			       (any_pkg pkg)#build_dependencies
			   | _  -> [] 
		       ) in
	   accumulate_deps add_package set' deps' followup_action
	)
	pkg_list
    )
  in

  (* Add the packages of [selected] in turn: *)
  let set =
    try
      accumulate_deps 
	add_package
	StrMap.empty
	(List.map 
	   (fun pkg -> parse_dependency ( (any_pkg pkg)#name ^ "-*" ) )
	   selected
	)
	(fun x -> x) 
    with
	Dead_end ->
	  (* It is not possible to fulfill all conditions! *)
	  ( match !msg with
		Some m -> raise(Mission_impossible m)
	      | None   -> raise(Mission_impossible "Unknown reason")
	  )
  in
  (* Make a list from the set: *)
  StrMap.fold
    (fun name (pkg,_) l -> pkg::l)
    set
    []
;;


let get_dependent_packages ~active ~selected () =
  let actv_ht = create_pkg_hash active in
  let succ_ht = Hashtbl.create 100 in
      (* maps name of package to list of successors *)
  let visited_ht = Hashtbl.create 100 in
      (* a package is "visited" if all successors are known *)

  let init_ht() =
    List.iter
      (fun pkg ->
	 let p = any_pkg pkg in
	 let name = p # name in
	 let deps = p # dependencies in
	 List.iter
	   (fun (dep_name,_) ->
	      let l = 
		try Hashtbl.find succ_ht dep_name 
		with Not_found -> StrSet.empty in
	      if not (StrSet.mem name l) then (
		Hashtbl.replace succ_ht dep_name (StrSet.add name l)
	      )
	   )
	   deps
      )
      active
  in

  let rec traverse pkg =
    let p = any_pkg pkg in
    let name = p # name in

    if not (Hashtbl.mem visited_ht pkg) then (

      let succ_set =
	try 
	  Hashtbl.find succ_ht name
	with 
	    Not_found -> StrSet.empty in
      
      let succ_names =
	StrSet.elements succ_set in
      
      let succ_pkgs =
	List.flatten
	  (List.map
	     (fun n ->
		let dep = Godi_versions.any_dependency n in
		find_pkg actv_ht n [ dep ])
	     succ_names
	  ) in

      Hashtbl.replace visited_ht pkg ();

      List.iter traverse succ_pkgs;

      let all_succs =
	List.fold_left
	  (fun acc n ->
	     let s = 
	       try Hashtbl.find succ_ht n with Not_found -> StrSet.empty  in
	     StrSet.union acc s
	  )
	  succ_set
	  succ_names in
      Hashtbl.replace succ_ht name all_succs;
    )
  in

  (* Initialize succ_ht with the direct dependencies: *)
  init_ht();
  (* Find the closure of succ_ht: *)
  List.iter traverse active;
  (* Return the result, which is contained in succ_ht: *)
  let all_succs =
    List.fold_left
      (fun acc pkg ->
	 let p = any_pkg pkg in
	 let succs =
	   try Hashtbl.find succ_ht p#name with Not_found -> StrSet.empty in
	 StrSet.union acc succs)
      StrSet.empty
      selected in
  List.flatten
    (List.map
       (fun name ->
	  let dep = Godi_versions.any_dependency name in
	  find_pkg actv_ht name [ dep ])
       (StrSet.elements all_succs)
    )
;;


let sort_packages pkgs =
  let pkgs_ht = create_pkg_hash pkgs in
  let visited = Hashtbl.create 100 in

  let rec antecedents pkg =
    if not(Hashtbl.mem visited pkg) then (
      let deps = (any_pkg pkg)#dependencies @
		 ( match pkg with
		       `Source_pkg _ -> 
			 (any_pkg pkg)#build_dependencies
		     | _  -> [] 
		 ) in
      Hashtbl.replace visited pkg ();
      List.fold_left
	(fun acc (name,cond) ->
	   let dep_pkgs = find_pkg pkgs_ht name [name,cond] in
	   let dep_pkgs' =
	     List.fold_left
	       (fun acc' dep_pkg ->
		  let dep_pkgs' = antecedents dep_pkg in
		  acc' @ dep_pkgs'
	       )
	       []
	       dep_pkgs in
	   acc @ dep_pkgs'
	)
	[]
	deps
      @ [ pkg ]
    )
    else []
  in

  List.fold_left
    (fun acc pkg ->
       let pkg' = antecedents pkg in
       acc @ pkg'
    )
    []
    pkgs
;;


let make_set l =
  PkgMap.make (List.map (fun x -> x, ()) l)
;;


let update_plan ~iteration pkgs =
  let count = ref 0 in
  let iteration_ht = Hashtbl.create 100 in

  let get_build_iteration pkg =
    try
      match Hashtbl.find iteration_ht pkg with
	  `B i -> i
	| _    -> raise Not_found
      with
	  Not_found -> iteration 
  in

  let get_remove_iteration pkg =
    try
      match Hashtbl.find iteration_ht pkg with
	  `R i -> i
	| _    -> raise Not_found
      with
	  Not_found -> iteration 
  in

  (* First set all `Remove_dep and `Build_dep packages to `Keep.
   * We remember the old iteration counter (iteration_ht).
   *)
  List.iter
    (fun pkg ->
       let p = any_pkg pkg in
       match p#user_status with
	   `Remove_dep i ->
	     Hashtbl.add iteration_ht pkg (`R i);
	     p#set_user_status `Keep
	 | `Build_dep i ->
	     Hashtbl.add iteration_ht pkg (`B i);
	     p#set_user_status `Keep
	 | _ ->
	     ()
    )
    pkgs;

  (* Partition pkgs: *)

  let inst_pkgs, pkgs' =
    List.partition
      (function `Installed_pkg _ -> true | _ -> false)
      pkgs in

  let src_pkgs, bin_pkgs =
    List.partition
      (function `Source_pkg _ -> true | _ -> false)
      pkgs' in

  let src_set = make_set src_pkgs in

  (* Compute the closure of all packages to remove: *)

  (* Get installed packages that need to be removed by deps: *)

  let rm_pkgs =
    List.filter to_be_removed inst_pkgs in

  let rm_inst_pkgs =
    get_dependent_packages ~active:inst_pkgs ~selected:rm_pkgs () in

  let rm_set =
    PkgMap.union (make_set rm_pkgs) (make_set rm_inst_pkgs) in

  List.iter
    (fun pkg ->
       let p = any_pkg pkg in
       if p # user_status = `Keep then (
	 let i = get_remove_iteration pkg in
	 p#set_user_status (`Remove_dep i);
	 if i = iteration then
	   incr count
       )
    )
    rm_inst_pkgs;

  (* After the removal, the set of installed packages is smaller: *)

  let inst_pkgs' = 
    List.filter (fun pkg -> not (PkgMap.mem pkg rm_set)) inst_pkgs in
  let inst_set' =
    make_set inst_pkgs' in

  (* Get installed packages that need to be rebuilt: 
   *
   * (~active is inst_pkgs, not inst_pkgs', because the question is which
   * packages were previously installed.)
   *)

  let inst_pkgs_to_keep =
    (* Installed packages the user wishes to keep *)
    List.filter
      (fun p -> pkg_user_status p <> `Remove)
      inst_pkgs in

  let build_pkgs =
    List.filter to_be_built src_pkgs in

  let sel_inst_pkgs =
    get_dependent_packages ~active:inst_pkgs_to_keep ~selected:build_pkgs () in

  (* Find the corresponding source packages, and select them: *)

  List.iter
    (fun pkg ->
       let name = (any_pkg pkg) # name in
       try
	 let src_pkg =
	   match PkgMap.find_by ~name src_set with
	     | []       -> raise Not_found
	     | [ p, _ ] -> p
	     | _        -> assert false in
	 let s = source_pkg src_pkg in
	 ( match s#user_status with
	     | `Keep ->
		 let i = get_build_iteration src_pkg in
		 s # set_user_status (`Build_dep i);
		 if i = iteration then
		   incr count
	     | _ ->
		 ()
	 )
       with
	   Not_found -> ()
    )
    sel_inst_pkgs;

  (* Do the requirements analysis: *)

  (* Get the list of non-installed source packages: *)

  let ninst_src_pkgs =
    List.filter
      (fun pkg ->
	 let name = (any_pkg pkg) # name in 
	 not (PkgMap.mem_by ~name inst_set')
      )
      src_pkgs in

  let selected = 
    List.filter
      (fun ip ->
	 let name = (any_pkg ip)#name in
	 let list = List.map fst (PkgMap.find_by ~name src_set) in
	 not(List.exists to_be_built list)
      )
      inst_pkgs'
    @
    List.filter to_be_built src_pkgs in

  (* Because packages to rebuild may occur both in inst_pkgs' and in
   * build_pkgs, the question is which object should take precedence.
   * Of course, the source package, as the dependencies may have changed.
   *)

   (* CHECK: The case that a newly selected package requires a new version
    * of an already selected package is not yet handled.
    *)

  let reqs =
    get_requirements
      ~available:ninst_src_pkgs
      ~selected
      () 
  in

  (* Select the missing packages: *)

  List.iter
    (fun req ->
       match req with
	   `Source_pkg p ->
	     (* We are only interested in missing source packages *)
	     ( match p#user_status with
		 | `Keep ->
		     let i = get_build_iteration req in
		     p # set_user_status (`Build_dep i);
		     if i = iteration then
		       incr count
		 | _ ->
		     ()
	     )
	 | _ ->
	     ()
    )
    reqs;

  (* Finally, schedule installed packages for removal, if the corresponding
   * source package will be built:
   *)

  List.iter
    (fun pkg ->
       let p = any_pkg pkg in
       let name = p#name in
       let list = List.map fst (PkgMap.find_by ~name src_set) in
       if List.exists to_be_built list then (
	 let i = get_remove_iteration pkg in
	 p#set_user_status (`Remove_dep i);
	 if i = iteration then
	   incr count
       )
    )
    inst_pkgs';

  !count
;;
